@@ -4,13 +4,13 @@ and the Massachusetts Institute of Technology, Cambridge, Massachusetts.
 
                         All Rights Reserved
 
-Permission to use, copy, modify, and distribute this software and its 
-documentation for any purpose and without fee is hereby granted, 
+Permission to use, copy, modify, and distribute this software and its
+documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
-both that copyright notice and this permission notice appear in 
+both that copyright notice and this permission notice appear in
 supporting documentation, and that the names of Digital or MIT not be
 used in advertising or publicity pertaining to distribution of the
-software without specific, written prior permission.  
+software without specific, written prior permission.
 
 DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
@@ -21,14 +21,12 @@ ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 SOFTWARE.
 
 ******************************************************************/
-/* $XFree86: xc/lib/Xv/Xv.c,v 1.17 2003/04/28 16:56:27 dawes Exp $ */
-/* $XdotOrg: xc/lib/Xv/Xv.c,v 1.2 2004/04/23 18:43:55 eich Exp $ */
 /*
-** File: 
+** File:
 **
 **   Xv.c --- Xv library extension module.
 **
-** Author: 
+** Author:
 **
 **   David Carver (Digital Workstation Engineering/Project Athena)
 **
@@ -51,22 +49,39 @@ SOFTWARE.
 **
 */
 
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+
 #include <stdio.h>
 #include "Xvlibint.h"
 #include <X11/extensions/Xext.h>
 #include <X11/extensions/extutil.h>
 #include <X11/extensions/XShm.h>
+#include <limits.h>
+
+#ifndef HAVE__XEATDATAWORDS
+static inline void _XEatDataWords(Display *dpy, unsigned long n)
+{
+# ifndef LONG64
+    if (n >= (ULONG_MAX >> 2))
+        _XIOError(dpy);
+# endif
+    _XEatData (dpy, n << 2);
+}
+#endif
 
 static XExtensionInfo _xv_info_data;
 static XExtensionInfo *xv_info = &_xv_info_data;
-static char *xv_extension_name = XvName;
+static const char *xv_extension_name = XvName;
 
 #define XvCheckExtension(dpy, i, val) \
   XextCheckExtension(dpy, i, xv_extension_name, val)
 
-static char *xv_error_string();
-static int xv_close_display();
-static Bool xv_wire_to_event();
+static char *xv_error_string(Display *dpy, int code, XExtCodes *codes,
+			     char * buf, int n);
+static int xv_close_display(Display *dpy, XExtCodes *codes);
+static Bool xv_wire_to_event(Display *dpy, XEvent *host, xEvent *wire);
 
 static XExtensionHooks xv_extension_hooks = {
     NULL,                               /* create_gc */
@@ -83,7 +98,7 @@ static XExtensionHooks xv_extension_hooks = {
 };
 
 
-static char *xv_error_list[] = 
+static const char *xv_error_list[] =
 {
    "BadPort",	    /* XvBadPort     */
    "BadEncoding",   /* XvBadEncoding */
@@ -93,11 +108,11 @@ static char *xv_error_list[] =
 static XEXT_GENERATE_CLOSE_DISPLAY (xv_close_display, xv_info)
 
 
-static XEXT_GENERATE_FIND_DISPLAY (xv_find_display, xv_info, 
-                                   xv_extension_name, 
+static XEXT_GENERATE_FIND_DISPLAY (xv_find_display, xv_info,
+                                   xv_extension_name,
                                    &xv_extension_hooks,
 				   XvNumEvents, NULL)
-     
+
 
 static XEXT_GENERATE_ERROR_STRING (xv_error_string, xv_extension_name,
                                    XvNumErrors, xv_error_list)
@@ -155,14 +170,14 @@ XvQueryAdaptors(
   XvAdaptorInfo *pas, *pa;
   XvFormat *pfs, *pf;
   char *buffer;
-  union 
+  union
     {
       char *buffer;
       char *string;
       xvAdaptorInfo *pa;
       xvFormat *pf;
     } u;
-  
+
   XvCheckExtension(dpy, info, XvBadExtension);
 
   LockDisplay(dpy);
@@ -255,7 +270,7 @@ XvQueryAdaptors(
 	  pf->depth = u.pf->depth;
 	  pf->visual_id = u.pf->visual;
 	  pf++;
-	  
+
 	  u.buffer += (sz_xvFormat + 3) & ~3;
       }
 
@@ -297,7 +312,7 @@ XvFreeAdaptorInfo(XvAdaptorInfo *pAdaptors)
 	{
 	  Xfree(pa->formats);
 	}
-    } 
+    }
 
   Xfree(pAdaptors);
 }
@@ -316,13 +331,13 @@ XvQueryEncodings(
   char *name;
   XvEncodingInfo *pes, *pe;
   char *buffer;
-  union 
+  union
     {
       char *buffer;
       char *string;
       xvEncodingInfo *pe;
     } u;
-  
+
   XvCheckExtension(dpy, info, XvBadExtension);
 
   LockDisplay(dpy);
@@ -380,6 +395,7 @@ XvQueryEncodings(
       u.buffer += (sz_xvEncodingInfo + 3) & ~3;
 
       if ( (name = (char *)Xmalloc(size+1)) == NULL) {
+	  XvFreeEncodingInfo(pes);
 	  Xfree(buffer);
           UnlockDisplay(dpy);
           SyncHandle();
@@ -427,7 +443,7 @@ XvPutVideo(
      XvPortID port,
      Drawable d,
      GC gc,
-     int vx, int vy, 
+     int vx, int vy,
      unsigned int vw, unsigned int vh,
      int dx, int dy,
      unsigned int dw, unsigned int dh
@@ -438,7 +454,7 @@ XvPutVideo(
   XvCheckExtension(dpy, info, XvBadExtension);
 
   LockDisplay(dpy);
-  
+
   FlushGC(dpy, gc);
 
   XvGetReq(PutVideo, req);
@@ -467,7 +483,7 @@ XvPutStill(
      XvPortID port,
      Drawable d,
      GC gc,
-     int vx, int vy, 
+     int vx, int vy,
      unsigned int vw, unsigned int vh,
      int dx, int dy,
      unsigned int dw, unsigned int dh
@@ -506,7 +522,7 @@ XvGetVideo(
      XvPortID port,
      Drawable d,
      GC gc,
-     int vx, int vy, 
+     int vx, int vy,
      unsigned int vw, unsigned int vh,
      int dx, int dy,
      unsigned int dw, unsigned int dh
@@ -545,7 +561,7 @@ XvGetStill(
      XvPortID port,
      Drawable d,
      GC gc,
-     int vx, int vy, 
+     int vx, int vy,
      unsigned int vw, unsigned int vh,
      int dx, int dy,
      unsigned int dw, unsigned int dh
@@ -620,7 +636,7 @@ XvGrabPort(
   req->port = port;
   req->time = time;
 
-  if (_XReply (dpy, (xReply *) &rep, 0, xTrue) == 0) 
+  if (_XReply (dpy, (xReply *) &rep, 0, xTrue) == 0)
     rep.result = GrabSuccess;
 
   result = rep.result;
@@ -755,7 +771,7 @@ XvGetPortAttribute (
   }
 
   *p_value = rep.value;
-  
+
   UnlockDisplay(dpy);
   SyncHandle();
 
@@ -767,11 +783,11 @@ XvQueryBestSize(
      Display *dpy,
      XvPortID port,
      Bool motion,
-     unsigned int vid_w, 
+     unsigned int vid_w,
      unsigned int vid_h,
-     unsigned int drw_w, 
+     unsigned int drw_w,
      unsigned int drw_h,
-     unsigned int *p_actual_width, 
+     unsigned int *p_actual_width,
      unsigned int *p_actual_height
 )
 {
@@ -809,7 +825,7 @@ XvQueryBestSize(
 }
 
 
-XvAttribute* 
+XvAttribute*
 XvQueryPortAttributes(Display *dpy, XvPortID port, int *num)
 {
   XExtDisplayInfo *info = xv_find_display(dpy);
@@ -834,26 +850,53 @@ XvQueryPortAttributes(Display *dpy, XvPortID port, int *num)
       return ret;
   }
 
-  if(rep.num_attributes) {
-      int size = (rep.num_attributes * sizeof(XvAttribute)) + rep.text_size;
+  /*
+   * X server sends data packed as:
+   *   attribute1, name1, attribute2, name2, ...
+   * We allocate a single buffer large enough to hold them all and
+   * then de-interleave the data so we return it to clients as:
+   *   attribute1, attribute2, ..., name1, name2, ...
+   * so that clients may refer to attributes as a simple array of
+   * structs:  attributes[0], attributes[1], ...
+   * and free it as a single/simple buffer.
+   */
 
-      if((ret = Xmalloc(size))) {
+  if(rep.num_attributes) {
+      unsigned long size;
+      /* limit each part to no more than one half the max size */
+      if ((rep.num_attributes < ((INT_MAX / 2) / sizeof(XvAttribute))) &&
+	  (rep.text_size < (INT_MAX / 2)-1)) {
+	  size = (rep.num_attributes * sizeof(XvAttribute)) + rep.text_size + 1;
+	  ret = Xmalloc(size);
+      }
+
+      if (ret != NULL) {
 	  char* marker = (char*)(&ret[rep.num_attributes]);
 	  xvAttributeInfo Info;
 	  int i;
-	
+
+	  /* keep track of remaining room for text strings */
+	  size = rep.text_size;
+
 	  for(i = 0; i < rep.num_attributes; i++) {
              _XRead(dpy, (char*)(&Info), sz_xvAttributeInfo);
-	      ret[i].flags = (int)Info.flags;	      
-	      ret[i].min_value = Info.min;	      
-	      ret[i].max_value = Info.max;	      
+	      ret[i].flags = (int)Info.flags;
+	      ret[i].min_value = Info.min;
+	      ret[i].max_value = Info.max;
 	      ret[i].name = marker;
-	      _XRead(dpy, marker, Info.size);
-	      marker += Info.size;
+	      if (Info.size <= size) {
+		  _XRead(dpy, marker, Info.size);
+		  marker += Info.size;
+		  size -= Info.size;
+	      }
 	      (*num)++;
 	  }
+
+	  /* ensure final string is nil-terminated to avoid exposure of
+             uninitialized memory */
+	  *marker = '\0';
       } else
-	_XEatData(dpy, rep.length << 2);
+	  _XEatDataWords(dpy, rep.length);
   }
 
   UnlockDisplay(dpy);
@@ -890,27 +933,28 @@ XvImageFormatValues * XvListImageFormats (
   }
 
   if(rep.num_formats) {
-      int size = (rep.num_formats * sizeof(XvImageFormatValues));
+      if (rep.num_formats < (INT_MAX / sizeof(XvImageFormatValues)))
+	  ret = Xmalloc(rep.num_formats * sizeof(XvImageFormatValues));
 
-      if((ret = Xmalloc(size))) {
+      if (ret != NULL) {
 	  xvImageFormatInfo Info;
 	  int i;
-	
+
 	  for(i = 0; i < rep.num_formats; i++) {
               _XRead(dpy, (char*)(&Info), sz_xvImageFormatInfo);
-	      ret[i].id = Info.id;	      
-	      ret[i].type = Info.type;	      
-	      ret[i].byte_order = Info.byte_order;	      
+	      ret[i].id = Info.id;
+	      ret[i].type = Info.type;
+	      ret[i].byte_order = Info.byte_order;
 	      memcpy(&(ret[i].guid[0]), &(Info.guid[0]), 16);
-	      ret[i].bits_per_pixel = Info.bpp;	      
-  	      ret[i].format = Info.format;	      
-   	      ret[i].num_planes = Info.num_planes;	      
-    	      ret[i].depth = Info.depth;	      
-    	      ret[i].red_mask = Info.red_mask;	      
-    	      ret[i].green_mask = Info.green_mask;	      
-    	      ret[i].blue_mask = Info.blue_mask;	      
-    	      ret[i].y_sample_bits = Info.y_sample_bits;	      
-    	      ret[i].u_sample_bits = Info.u_sample_bits;	      
+	      ret[i].bits_per_pixel = Info.bpp;
+  	      ret[i].format = Info.format;
+   	      ret[i].num_planes = Info.num_planes;
+    	      ret[i].depth = Info.depth;
+    	      ret[i].red_mask = Info.red_mask;
+    	      ret[i].green_mask = Info.green_mask;
+    	      ret[i].blue_mask = Info.blue_mask;
+    	      ret[i].y_sample_bits = Info.y_sample_bits;
+    	      ret[i].u_sample_bits = Info.u_sample_bits;
     	      ret[i].v_sample_bits = Info.v_sample_bits;
     	      ret[i].horz_y_period = Info.horz_y_period;
     	      ret[i].horz_u_period = Info.horz_u_period;
@@ -923,7 +967,7 @@ XvImageFormatValues * XvListImageFormats (
 	      (*num)++;
 	  }
       } else
-	_XEatData(dpy, rep.length << 2);
+	  _XEatDataWords(dpy, rep.length);
   }
 
   UnlockDisplay(dpy);
@@ -937,8 +981,8 @@ XvImage * XvCreateImage (
    XvPortID port,
    int id,
    char *data,
-   int width, 
-   int height 
+   int width,
+   int height
 ) {
    XExtDisplayInfo *info = xv_find_display(dpy);
    xvQueryImageAttributesReq *req;
@@ -963,7 +1007,10 @@ XvImage * XvCreateImage (
       return NULL;
    }
 
-   if((ret = (XvImage*)Xmalloc(sizeof(XvImage) + (rep.num_planes << 3)))) {
+   if (rep.num_planes < ((INT_MAX >> 3) - sizeof(XvImage)))
+       ret = Xmalloc(sizeof(XvImage) + (rep.num_planes << 3));
+
+   if (ret != NULL) {
 	ret->id = id;
 	ret->width = rep.width;
 	ret->height = rep.height;
@@ -976,7 +1023,7 @@ XvImage * XvCreateImage (
   	_XRead(dpy, (char*)(ret->pitches), rep.num_planes << 2);
 	_XRead(dpy, (char*)(ret->offsets), rep.num_planes << 2);
    } else
-	_XEatData(dpy, rep.length << 2);
+       _XEatDataWords(dpy, rep.length);
 
    UnlockDisplay(dpy);
    SyncHandle();
@@ -989,7 +1036,7 @@ XvImage * XvShmCreateImage (
    XvPortID port,
    int id,
    char *data,
-   int width, 
+   int width,
    int height,
    XShmSegmentInfo *shminfo
 ){
@@ -1012,7 +1059,7 @@ int XvPutImage (
    int src_y,
    unsigned int src_w,
    unsigned int src_h,
-   int dest_x, 
+   int dest_x,
    int dest_y,
    unsigned int dest_w,
    unsigned int dest_h
@@ -1049,7 +1096,7 @@ int XvPutImage (
 
   /* Yes it's kindof lame that we are sending the whole thing,
      but for video all of it may be needed even if displaying
-     only a subsection, and I don't want to go through the 
+     only a subsection, and I don't want to go through the
      trouble of creating subregions to send */
   Data(dpy, (char *)image->data, image->data_size);
 
@@ -1069,7 +1116,7 @@ int XvShmPutImage (
    int src_y,
    unsigned int src_w,
    unsigned int src_h,
-   int dest_x, 
+   int dest_x,
    int dest_y,
    unsigned int dest_w,
    unsigned int dest_h,
@@ -1082,7 +1129,7 @@ int XvShmPutImage (
   XvCheckExtension(dpy, info, XvBadExtension);
 
   LockDisplay(dpy);
-  
+
   FlushGC(dpy, gc);
 
   XvGetReq(ShmPutImage, req);
@@ -1125,7 +1172,7 @@ xv_wire_to_event(Display *dpy, XEvent *host, xEvent *wire)
   {
     case XvVideoNotify:
       re->xvvideo.type = event->u.u.type & 0x7f;
-      re->xvvideo.serial = 
+      re->xvvideo.serial =
 	_XSetLastRequestRead(dpy, (xGenericReply *)event);
       re->xvvideo.send_event = ((event->u.u.type & 0x80) != 0);
       re->xvvideo.display = dpy;
@@ -1136,7 +1183,7 @@ xv_wire_to_event(Display *dpy, XEvent *host, xEvent *wire)
       break;
     case XvPortNotify:
       re->xvport.type = event->u.u.type & 0x7f;
-      re->xvport.serial = 
+      re->xvport.serial =
 	_XSetLastRequestRead(dpy, (xGenericReply *)event);
       re->xvport.send_event = ((event->u.u.type & 0x80) != 0);
       re->xvport.display = dpy;
@@ -1146,7 +1193,7 @@ xv_wire_to_event(Display *dpy, XEvent *host, xEvent *wire)
       re->xvport.value = event->u.portNotify.value;
       break;
     default:
-      return False; 
+      return False;
   }
 
   return (True);
